@@ -12,7 +12,7 @@
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * RCS: @(#) $Id: tclEnv.c,v 1.7.2.5 2002/10/15 20:23:51 hobbs Exp $
+ * RCS: @(#) $Id: tclEnv.c,v 1.20.2.4 2007/08/07 05:04:48 das Exp $
  */
 
 #include "tclInt.h"
@@ -26,6 +26,11 @@ static char **environCache = NULL;
 				 * strings that Tcl has allocated. */
 
 #ifndef USE_PUTENV
+static char **ourEnviron = NULL;/* Cache of the array that we allocate.
+				 * We need to track this in case another
+				 * subsystem swaps around the environ array
+				 * like we do.
+				 */
 static int environSize = 0;	/* Non-zero means that the environ array was
 				 * malloced and has this many total entries
 				 * allocated to it (not all may be in use at
@@ -33,21 +38,13 @@ static int environSize = 0;	/* Non-zero means that the environ array was
 				 * array is in its original static state. */
 #endif
 
-/*
- * For MacOS X
- */
-#if defined(__APPLE__) && defined(__DYNAMIC__)
-#include <crt_externs.h>
-char **environ = NULL;
-#endif
-
 /*
  * Declarations for local procedures defined in this file:
  */
 
 static char *		EnvTraceProc _ANSI_ARGS_((ClientData clientData,
-			    Tcl_Interp *interp, char *name1, char *name2,
-			    int flags));
+			    Tcl_Interp *interp, CONST char *name1, 
+			    CONST char *name2, int flags));
 static void		ReplaceString _ANSI_ARGS_((CONST char *oldStr,
 			    char *newStr));
 void			TclSetEnv _ANSI_ARGS_((CONST char *name,
@@ -90,13 +87,6 @@ TclSetupEnv(interp)
     char *p1, *p2;
     int i;
 
-    /*
-     * For MacOS X
-     */
-#if defined(__APPLE__) && defined(__DYNAMIC__)
-    environ = *_NSGetEnviron();
-#endif
-
     /*
      * Synchronize the values in the environ array with the contents
      * of the Tcl "env" variable.  To do this:
@@ -177,7 +167,8 @@ TclSetEnv(name, value)
 {
     Tcl_DString envString;
     int index, length, nameLength;
-    char *p, *p2, *oldValue;
+    char *p, *oldValue;
+    CONST char *p2;
 
     /*
      * Figure out where the entry is going to go.  If the name doesn't
@@ -190,24 +181,23 @@ TclSetEnv(name, value)
 
     if (index == -1) {
 #ifndef USE_PUTENV
-	if ((length + 2) > environSize) {
+	/*
+	 * We need to handle the case where the environment may be changed
+	 * outside our control.  environSize is only valid if the current
+	 * environment is the one we allocated. [Bug 979640]
+	 */
+	if ((ourEnviron != environ) || ((length + 2) > environSize)) {
 	    char **newEnviron;
 
 	    newEnviron = (char **) ckalloc((unsigned)
 		    ((length + 5) * sizeof(char *)));
 	    memcpy((VOID *) newEnviron, (VOID *) environ,
 		    length*sizeof(char *));
-	    if (environSize != 0) {
-		ckfree((char *) environ);
+	    if ((environSize != 0) && (ourEnviron != NULL)) {
+		ckfree((char *) ourEnviron);
 	    }
-	    environ = newEnviron;
+	    environ = ourEnviron = newEnviron;
 	    environSize = length + 5;
-#if defined(__APPLE__) && defined(__DYNAMIC__)
-	    {
-	    char ***e = _NSGetEnviron();
-	    *e = environ;
-	    }
-#endif
 	}
 	index = length;
 	environ[index + 1] = NULL;
@@ -215,7 +205,7 @@ TclSetEnv(name, value)
 	oldValue = NULL;
 	nameLength = strlen(name);
     } else {
-	char *env;
+	CONST char *env;
 
 	/*
 	 * Compare the new value to the existing value.  If they're
@@ -236,7 +226,6 @@ TclSetEnv(name, value)
 	oldValue = environ[index];
 	nameLength = length;
     }
-	
 
     /*
      * Create a new entry.  Build a complete UTF string that contains
@@ -277,9 +266,23 @@ TclSetEnv(name, value)
 
     if ((index != -1) && (environ[index] == p)) {
 	ReplaceString(oldValue, p);
+#ifdef HAVE_PUTENV_THAT_COPIES
+    } else {
+	/* This putenv() copies instead of taking ownership */
+	ckfree(p);
+#endif
     }
 
     Tcl_MutexUnlock(&envMutex);
+    
+    if (!strcmp(name, "HOME")) {
+	/* 
+	 * If the user's home directory has changed, we must invalidate
+	 * the filesystem cache, because '~' expansions will now be
+	 * incorrect.
+	 */
+        Tcl_FSMountsChanged(NULL);
+    }
 }
 
 /*
@@ -311,7 +314,8 @@ Tcl_PutEnv(string)
 				 * form NAME=value. (native) */
 {
     Tcl_DString nameString;   
-    char *name, *value;
+    CONST char *name;
+    char *value;
 
     if (string == NULL) {
 	return 0;
@@ -362,7 +366,7 @@ TclUnsetEnv(name)
     char *oldValue;
     int length;
     int index;
-#ifdef USE_PUTENV
+#ifdef USE_PUTENV_FOR_UNSET
     Tcl_DString envString;
     char *string;
 #else
@@ -376,7 +380,7 @@ TclUnsetEnv(name)
      * First make sure that the environment variable exists to avoid
      * doing needless work and to avoid recursion on the unset.
      */
-    
+
     if (index == -1) {
 	Tcl_MutexUnlock(&envMutex);
 	return;
@@ -392,12 +396,22 @@ TclUnsetEnv(name)
      * update the interpreters or we will recurse.
      */
 
-#ifdef USE_PUTENV
-    string = ckalloc(length+2);
+#ifdef USE_PUTENV_FOR_UNSET
+    /*
+     * For those platforms that support putenv to unset, Linux indicates
+     * that no = should be included, and Windows requires it.
+     */
+#ifdef WIN32
+    string = ckalloc((unsigned int) length+2);
     memcpy((VOID *) string, (VOID *) name, (size_t) length);
     string[length] = '=';
     string[length+1] = '\0';
-    
+#else
+    string = ckalloc((unsigned int) length+1);
+    memcpy((VOID *) string, (VOID *) name, (size_t) length);
+    string[length] = '\0';
+#endif
+
     Tcl_UtfToExternalDString(NULL, string, -1, &envString);
     string = ckrealloc(string, (unsigned) (Tcl_DStringLength(&envString)+1));
     strcpy(string, Tcl_DStringValue(&envString));
@@ -413,6 +427,11 @@ TclUnsetEnv(name)
 
     if (environ[index] == string) {
 	ReplaceString(oldValue, string);
+#ifdef HAVE_PUTENV_THAT_COPIES
+    } else {
+	/* This putenv() copies instead of taking ownership */
+	ckfree(string);
+#endif
     }
 #else
     for (envPtr = environ+index+1; ; envPtr++) {
@@ -447,7 +466,7 @@ TclUnsetEnv(name)
  *----------------------------------------------------------------------
  */
 
-char *
+CONST char *
 TclGetEnv(name, valuePtr)
     CONST char *name;		/* Name of environment variable to find
 				 * (UTF-8). */
@@ -456,7 +475,7 @@ TclGetEnv(name, valuePtr)
 				 * stored. */
 {
     int length, index;
-    char *result;
+    CONST char *result;
 
     Tcl_MutexLock(&envMutex);
     index = TclpFindVariable(name, &length);
@@ -507,8 +526,8 @@ EnvTraceProc(clientData, interp, name1, name2, flags)
     ClientData clientData;	/* Not used. */
     Tcl_Interp *interp;		/* Interpreter whose "env" variable is
 				 * being modified. */
-    char *name1;		/* Better be "env". */
-    char *name2;		/* Name of variable being modified, or NULL
+    CONST char *name1;		/* Better be "env". */
+    CONST char *name2;		/* Name of variable being modified, or NULL
 				 * if whole array is being deleted (UTF-8). */
     int flags;			/* Indicates what's happening. */
 {
@@ -534,7 +553,7 @@ EnvTraceProc(clientData, interp, name1, name2, flags)
      */
 
     if (flags & TCL_TRACE_WRITES) {
-	char *value;
+	CONST char *value;
 	
 	value = Tcl_GetVar2(interp, "env", name2, TCL_GLOBAL_ONLY);
 	TclSetEnv(name2, value);
@@ -546,7 +565,7 @@ EnvTraceProc(clientData, interp, name1, name2, flags)
 
     if (flags & TCL_TRACE_READS) {
 	Tcl_DString valueString;
-	char *value;
+	CONST char *value;
 
 	value = TclGetEnv(name2, &valueString);
 	if (value == NULL) {
@@ -612,7 +631,7 @@ ReplaceString(oldStr, newStr)
 	if (environCache[i]) {
 	    ckfree(environCache[i]);
 	}
-	    
+
 	if (newStr) {
 	    environCache[i] = newStr;
 	} else {
@@ -621,7 +640,7 @@ ReplaceString(oldStr, newStr)
 	    }
 	    environCache[cacheSize-1] = NULL;
 	}
-    } else {	
+    } else {
         int allocatedSize = (cacheSize + 5) * sizeof(char *);
 
 	/*
@@ -630,14 +649,14 @@ ReplaceString(oldStr, newStr)
 
 	newCache = (char **) ckalloc((unsigned) allocatedSize);
         (VOID *) memset(newCache, (int) 0, (size_t) allocatedSize);
-        
+
 	if (environCache) {
 	    memcpy((VOID *) newCache, (VOID *) environCache,
 		    (size_t) (cacheSize * sizeof(char*)));
 	    ckfree((char *) environCache);
 	}
 	environCache = newCache;
-	environCache[cacheSize] = (char *) newStr;
+	environCache[cacheSize] = newStr;
 	environCache[cacheSize+1] = NULL;
 	cacheSize += 5;
     }
