@@ -9,13 +9,43 @@
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * RCS: @(#) $Id: tclUnixTime.c,v 1.7 2000/01/14 22:15:52 ericm Exp $
+ * RCS: @(#) $Id: tclUnixTime.c,v 1.15.2.7 2008/04/14 17:34:17 kennykb Exp $
  */
 
 #include "tclInt.h"
 #include "tclPort.h"
+#include <locale.h>
 #define TM_YEAR_BASE 1900
 #define IsLeapYear(x)   ((x % 4 == 0) && (x % 100 != 0 || x % 400 == 0))
+
+/*
+ * TclpGetDate is coded to return a pointer to a 'struct tm'.  For
+ * thread safety, this structure must be in thread-specific data.
+ * The 'tmKey' variable is the key to this buffer.
+ */
+
+static Tcl_ThreadDataKey tmKey;
+typedef struct ThreadSpecificData {
+    struct tm gmtime_buf;
+    struct tm localtime_buf;
+} ThreadSpecificData;
+
+/*
+ * If we fall back on the thread-unsafe versions of gmtime and localtime,
+ * use this mutex to try to protect them.
+ */
+
+TCL_DECLARE_MUTEX(tmMutex)
+
+static char* lastTZ = NULL;	/* Holds the last setting of the
+				 * TZ environment variable, or an
+				 * empty string if the variable was
+				 * not set. */
+
+/* Static functions declared in this file */
+
+static void SetTZIfNecessary _ANSI_ARGS_((void));
+static void CleanupMemory _ANSI_ARGS_((ClientData));
 
 /*
  *-----------------------------------------------------------------------------
@@ -67,13 +97,12 @@ TclpGetClicks()
     struct tms dummy;
 #else
     struct timeval date;
-    struct timezone tz;
 #endif
 
 #ifdef NO_GETTOD
     now = (unsigned long) times(&dummy);
 #else
-    gettimeofday(&date, &tz);
+    gettimeofday(&date, NULL);
     now = date.tv_sec*1000000 + date.tv_usec;
 #endif
 
@@ -101,21 +130,25 @@ TclpGetClicks()
 
 int
 TclpGetTimeZone (currentTime)
-    unsigned long  currentTime;
+    Tcl_WideInt  currentTime;
 {
     /*
-     * Determine how a timezone is obtained from "struct tm".  If there is no
-     * time zone in this struct (very lame) then use the timezone variable.
-     * This is done in a way to make the timezone variable the method of last
-     * resort, as some systems have it in addition to a field in "struct tm".
-     * The gettimeofday system call can also be used to determine the time
-     * zone.
+     * We prefer first to use the time zone in "struct tm" if the
+     * structure contains such a member.  Following that, we try
+     * to locate the external 'timezone' variable and use its value.
+     * If both of those methods fail, we attempt to convert a known
+     * time to local time and use the difference from UTC as the local
+     * time zone.  In all cases, we need to undo any Daylight Saving Time
+     * adjustment.
      */
     
 #if defined(HAVE_TM_TZADJ)
 #   define TCL_GOT_TIMEZONE
+
+    /* Struct tm contains tm_tzadj - that value may be used. */
+
     time_t      curTime = (time_t) currentTime;
-    struct tm  *timeDataPtr = localtime(&curTime);
+    struct tm  *timeDataPtr = TclpLocaltime((TclpTime_t) &curTime);
     int         timeZone;
 
     timeZone = timeDataPtr->tm_tzadj  / 60;
@@ -124,12 +157,16 @@ TclpGetTimeZone (currentTime)
     }
     
     return timeZone;
+
 #endif
 
 #if defined(HAVE_TM_GMTOFF) && !defined (TCL_GOT_TIMEZONE)
 #   define TCL_GOT_TIMEZONE
+
+    /* Struct tm contains tm_gmtoff - that value may be used. */
+
     time_t     curTime = (time_t) currentTime;
-    struct tm *timeDataPtr = localtime(&curTime);
+    struct tm *timeDataPtr = TclpLocaltime((TclpTime_t) &curTime);
     int        timeZone;
 
     timeZone = -(timeDataPtr->tm_gmtoff / 60);
@@ -138,46 +175,17 @@ TclpGetTimeZone (currentTime)
     }
     
     return timeZone;
+
 #endif
 
-#if defined(USE_DELTA_FOR_TZ)
-#define TCL_GOT_TIMEZONE 1
-    /*
-     * This hack replaces using global var timezone or gettimeofday
-     * in situations where they are buggy such as on AIX when libbsd.a
-     * is linked in.
-     */
-
-    int timeZone;
-    time_t tt;
-    struct tm *stm;
-    tt = 849268800L;      /*    1996-11-29 12:00:00  GMT */
-    stm = localtime(&tt); /* eg 1996-11-29  6:00:00  CST6CDT */
-    /* The calculation below assumes a max of +12 or -12 hours from GMT */
-    timeZone = (12 - stm->tm_hour)*60 + (0 - stm->tm_min);
-    return timeZone;  /* eg +360 for CST6CDT */
-#endif
-
-    /*
-     * Must prefer timezone variable over gettimeofday, as gettimeofday does
-     * not return timezone information on many systems that have moved this
-     * information outside of the kernel.
-     */
-    
-#if defined(HAVE_TIMEZONE_VAR) && !defined (TCL_GOT_TIMEZONE)
+#if defined(HAVE_TIMEZONE_VAR) && !defined(TCL_GOT_TIMEZONE) && !defined(USE_DELTA_FOR_TZ)
 #   define TCL_GOT_TIMEZONE
-    static int setTZ = 0;
-#ifdef TCL_THREADS
-    static Tcl_Mutex tzMutex;
-#endif
+
     int        timeZone;
 
-    Tcl_MutexLock(&tzMutex);
-    if (!setTZ) {
-        tzset();
-        setTZ = 1;
-    }
-    Tcl_MutexUnlock(&tzMutex);
+    /* The 'timezone' external var is present and may be used. */
+
+    SetTZIfNecessary();
 
     /*
      * Note: this is not a typo in "timezone" below!  See tzset
@@ -185,30 +193,36 @@ TclpGetTimeZone (currentTime)
      */
 
     timeZone = timezone / 60;
-
     return timeZone;
+
 #endif
 
-#if !defined(NO_GETTOD) && !defined (TCL_GOT_TIMEZONE)
-#   define TCL_GOT_TIMEZONE
-    struct timeval  tv;
-    struct timezone tz;
-    int timeZone;
+#if !defined(TCL_GOT_TIMEZONE) 
+#define TCL_GOT_TIMEZONE 1
+    /*
+     * Fallback - determine time zone with a known reference time.
+     */
 
-    gettimeofday(&tv, &tz);
-    timeZone = tz.tz_minuteswest;
-    if (tz.tz_dsttime) {
+    int timeZone;
+    time_t tt;
+    struct tm *stm;
+    tt = 849268800L;      /*    1996-11-29 12:00:00  GMT */
+    stm = TclpLocaltime((TclpTime_t) &tt); /* eg 1996-11-29  6:00:00  CST6CDT */
+    /* The calculation below assumes a max of +12 or -12 hours from GMT */
+    timeZone = (12 - stm->tm_hour)*60 + (0 - stm->tm_min);
+    if ( stm -> tm_isdst ) {
         timeZone += 60;
     }
-    
-    return timeZone;
+    return timeZone;  /* eg +360 for CST6CDT */
 #endif
 
 #ifndef TCL_GOT_TIMEZONE
     /*
      * Cause compile error, we don't know how to get timezone.
      */
-    error: autoconf did not figure out how to determine the timezone. 
+
+#error autoconf did not figure out how to determine the timezone. 
+
 #endif
 
 }
@@ -216,7 +230,7 @@ TclpGetTimeZone (currentTime)
 /*
  *----------------------------------------------------------------------
  *
- * TclpGetTime --
+ * Tcl_GetTime --
  *
  *	Gets the current system time in seconds and microseconds
  *	since the beginning of the epoch: 00:00 UCT, January 1, 1970.
@@ -231,13 +245,12 @@ TclpGetTimeZone (currentTime)
  */
 
 void
-TclpGetTime(timePtr)
+Tcl_GetTime(timePtr)
     Tcl_Time *timePtr;		/* Location to store time information. */
 {
     struct timeval tv;
-    struct timezone tz;
     
-    (void) gettimeofday(&tv, &tz);
+    (void) gettimeofday(&tv, NULL);
     timePtr->sec = tv.tv_sec;
     timePtr->usec = tv.tv_usec;
 }
@@ -265,12 +278,10 @@ TclpGetDate(time, useGMT)
     TclpTime_t time;
     int useGMT;
 {
-    CONST time_t *tp = (CONST time_t *)time;
-
     if (useGMT) {
-	return gmtime(tp);
+	return TclpGmtime(time);
     } else {
-	return localtime(tp);
+	return TclpLocaltime(time);
     }
 }
 
@@ -279,7 +290,8 @@ TclpGetDate(time, useGMT)
  *
  * TclpStrftime --
  *
- *	On Unix, we can safely call the native strftime implementation.
+ *	On Unix, we can safely call the native strftime implementation,
+ *	and also ignore the useGMT parameter.
  *
  * Results:
  *	The normal strftime result.
@@ -291,11 +303,12 @@ TclpGetDate(time, useGMT)
  */
 
 size_t
-TclpStrftime(s, maxsize, format, t)
+TclpStrftime(s, maxsize, format, t, useGMT)
     char *s;
     size_t maxsize;
     CONST char *format;
     CONST struct tm *t;
+    int useGMT;
 {
     if (format[0] == '%' && format[1] == 'Q') {
 	/* Format as a stardate */
@@ -306,5 +319,172 @@ TclpStrftime(s, maxsize, format, t)
 		(((t->tm_hour * 60) + t->tm_min)/144));
 	return(strlen(s));
     }
+    setlocale(LC_TIME, "");
     return strftime(s, maxsize, format, t);
 }
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * TclpGmtime --
+ *
+ *	Wrapper around the 'gmtime' library function to make it thread
+ *	safe.
+ *
+ * Results:
+ *	Returns a pointer to a 'struct tm' in thread-specific data.
+ *
+ * Side effects:
+ *	Invokes gmtime or gmtime_r as appropriate.
+ *
+ *----------------------------------------------------------------------
+ */
+
+struct tm *
+TclpGmtime( tt )
+    TclpTime_t_CONST tt;
+{
+    CONST time_t *timePtr = (CONST time_t *) tt;
+				/* Pointer to the number of seconds
+				 * since the local system's epoch */
+
+    /*
+     * Get a thread-local buffer to hold the returned time.
+     */
+
+    ThreadSpecificData *tsdPtr = TCL_TSD_INIT( &tmKey );
+#ifdef HAVE_GMTIME_R
+    gmtime_r(timePtr, &( tsdPtr->gmtime_buf ));
+#else
+    Tcl_MutexLock( &tmMutex );
+    memcpy( (VOID *) &( tsdPtr->gmtime_buf ),
+	    (VOID *) gmtime( timePtr ),
+	    sizeof( struct tm ) );
+    Tcl_MutexUnlock( &tmMutex );
+#endif    
+    return &( tsdPtr->gmtime_buf );
+}
+/*
+ * Forwarder for obsolete item in Stubs
+ */
+struct tm*
+TclpGmtime_unix( timePtr )
+    TclpTime_t_CONST timePtr;
+{
+    return TclpGmtime( timePtr );
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * TclpLocaltime --
+ *
+ *	Wrapper around the 'localtime' library function to make it thread
+ *	safe.
+ *
+ * Results:
+ *	Returns a pointer to a 'struct tm' in thread-specific data.
+ *
+ * Side effects:
+ *	Invokes localtime or localtime_r as appropriate.
+ *
+ *----------------------------------------------------------------------
+ */
+
+struct tm *
+TclpLocaltime( tt )
+    TclpTime_t_CONST tt;
+{
+    CONST time_t *timePtr = (CONST time_t *) tt;
+				/* Pointer to the number of seconds
+				 * since the local system's epoch */
+    /*
+     * Get a thread-local buffer to hold the returned time.
+     */
+
+    ThreadSpecificData *tsdPtr = TCL_TSD_INIT( &tmKey );
+    SetTZIfNecessary();
+#ifdef HAVE_LOCALTIME_R
+    localtime_r( timePtr, &( tsdPtr->localtime_buf ) );
+#else
+    Tcl_MutexLock( &tmMutex );
+    memcpy( (VOID *) &( tsdPtr -> localtime_buf ),
+	    (VOID *) localtime( timePtr ),
+	    sizeof( struct tm ) );
+    Tcl_MutexUnlock( &tmMutex );
+#endif    
+    return &( tsdPtr->localtime_buf );
+}
+/*
+ * Forwarder for obsolete item in Stubs
+ */
+struct tm*
+TclpLocaltime_unix( timePtr )
+    TclpTime_t_CONST timePtr;
+{
+    return TclpLocaltime( timePtr );
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * SetTZIfNecessary --
+ *
+ *	Determines whether a call to 'tzset' is needed prior to the
+ *	next call to 'localtime' or examination of the 'timezone' variable.
+ *
+ * Results:
+ *	None.
+ *
+ * Side effects:
+ *	If 'tzset' has never been called in the current process, or if
+ *	the value of the environment variable TZ has changed since the
+ *	last call to 'tzset', then 'tzset' is called again.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static void
+SetTZIfNecessary() {
+
+    CONST char* newTZ = getenv( "TZ" );
+    Tcl_MutexLock(&tmMutex);
+    if ( newTZ == NULL ) {
+	newTZ = "";
+    }
+    if ( lastTZ == NULL || strcmp( lastTZ, newTZ ) ) {
+        tzset();
+	if ( lastTZ == NULL ) {
+	    Tcl_CreateExitHandler( CleanupMemory, (ClientData) NULL );
+	} else {
+	    Tcl_Free( lastTZ );
+	}
+	lastTZ = ckalloc( strlen( newTZ ) + 1 );
+	strcpy( lastTZ, newTZ );
+    }
+    Tcl_MutexUnlock(&tmMutex);
+
+}
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * CleanupMemory --
+ *
+ *	Releases the private copy of the TZ environment variable
+ *	upon exit from Tcl.
+ *
+ * Results:
+ *	None.
+ *
+ * Side effects:
+ *	Frees allocated memory.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static void
+CleanupMemory( ClientData ignored )
+{
+    ckfree( lastTZ );
+}
